#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<omp.h>

void rot_1(int N, double* a) {
  double tmp = a[0];
  #pragma omp parallel
  {
    int last = 0;
    #pragma omp for schedule(static) nowait
    for (int i = 0; i < N - 1; i++) {
      last = i;
    }
    double a_last = a[last+1];
    #pragma omp barrier
    #pragma omp for schedule(static) nowait
    for (int i = 0; i < N - 1; i++) {
      a[i] = a[i+1];
    }
    a[last] = a_last;
  }
  a[N - 1] = tmp;
}

void rot(int N, double* a) {
  double tmp = a[0];
  #pragma omp parallel
  {
    int last = 0;
    int nthr = omp_get_num_threads();
    int tid = omp_get_thread_num();
    int ichunk = (N - 1) / nthr;
    int ifrom = ichunk * tid;
    int ito = ifrom + ichunk;
    if (tid == nthr - 1) ito = N - 1;
    last = ito - 1;
    double a_last = a[last+1];
    #pragma omp barrier
    for (int i = ifrom; i < ito; i++) {
      a[i] = a[i+1];
    }
    a[last] = a_last;
  }
  a[N - 1] = tmp;
}

int main(int argc, char ** argv) {
  if (argc != 2) {
    printf("Usage: %s <N>\n", argv[0]);
    return EXIT_SUCCESS;
  }
  int N = atoi(argv[1]);
  double* a = malloc(N * sizeof(double));
  for (int i = 0; i < N; i++) a[i] = i;
  double start = omp_get_wtime();
  rot(N, a);
  double end = omp_get_wtime();
  double error = 0;
  for (int i = 0; i < N; i++) error += fabs(a[i] - (i + 1) % N);
  printf("elapsed: %f, error: %f\n", end - start, error);
}
