#include<stdio.h>
#include<stdlib.h>
#include<omp.h>
#include<math.h>

typedef struct tree_t {
  struct tree_t *left, *right;
  double data;
  double annotation;
} tree_t;

tree_t* build(int n, int max) {
  if (n > max) return NULL;
  tree_t* result = malloc(sizeof(tree_t));
  result->left = build(2 * n, max);
  result->right = build(2 * n + 1, max);
  result->data = n * (double) n;
  result->annotation = 0;
  return result;
}

void annotate(tree_t* tree) {
  if (tree->left)
    #pragma omp task if(tree->data < 1000)
    annotate(tree->left);
  if (tree->right) 
    #pragma omp task if(tree->data < 1000)
    annotate(tree->right);
  tree->annotation = sqrt(tree->data);
}

double compute(tree_t* tree) {
  double l = 0, r = 0;
  if (tree->left) 
    #pragma omp task shared(l) if(tree->data < 1000)
    l = compute(tree->left);
  if (tree->right) 
    #pragma omp task shared(r) if(tree->data < 1000)
    r = compute(tree->right);
  #pragma omp taskwait
  return l + r + tree->annotation;
}

double execute(tree_t* tree) {
  double result;
  #pragma omp parallel
  #pragma omp single
  {
  #pragma omp taskgroup
  {
  annotate(tree);
  }
  result = compute(tree);
  }
  return result;
}

int main(int argc, char ** argv) {
  if (argc != 2) {
    printf("Usage: %s <N>\n", argv[0]);
    return EXIT_SUCCESS;
  }
  int N = atoi(argv[1]) + 1;
  tree_t * tree = build(1, N);
  double start = omp_get_wtime();
  double result = execute(tree);
  double end = omp_get_wtime();
  double error = fabs(result - N * (double) (N + 1) / (double) 2);
  printf("elapsed: %f, error: %f (%f)\n", end - start, error, result);
}
